/* This file is part of mfmod_openmetrics.
 * Copyright (C) 2022 Sergey Poznyakoff
 *
 * Mfmod_openmetrics is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * Mfmod_openmetrics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mfmod_openmetrics.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"
#include <mailfromd/mfmod.h>
#include <mailfromd/exceptions.h>
#include <mailutils/mailutils.h>
#include <gdbm.h>
#include <inttypes.h>
#include <microhttpd.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netdb.h>

static int
enumargs(MFMOD_PARAM *r)
{
	r->type = mfmod_string;
	r->string = strdup("bad number of arguments");
	return mfe_inval;
}

static char const *
mfmod_type_str(int t)
{
	static char const *type_str[] = {
		[mfmod_number] = "number",
		[mfmod_string] = "string",
		[mfmod_message] = "message"
	};
	if (t >= 0 && t < sizeof(type_str)/sizeof(type_str[0]) && type_str[t])
		return type_str[t];
	return "UNKNOWN";
}

static int
eargtype(MFMOD_PARAM *r, int n, int exptype, MFMOD_PARAM *p)
{
	char errbuf[128];

	snprintf(errbuf, sizeof(errbuf),
		 "bad type of argument #%d: expected %s, but given %s",
		 n+1, mfmod_type_str(exptype), mfmod_type_str(p->type));
	r->type = mfmod_string;
	r->string = strdup(errbuf);
	return mfe_inval;
}

static int
egdbm(MFMOD_PARAM *r)
{
	r->type = mfmod_string;
	r->string = mu_strdup(gdbm_strerror(gdbm_errno));
	return mfe_dbfailure;
}

#define ASSERT_ARGTYPE(p, r, n, t)                                      \
	do {                                                            \
		if (p[n].type != t) return eargtype(r, n, t, &p[n]);    \
	} while (0)

typedef uint64_t COUNTER;

typedef struct {
	size_t off_name;
	COUNTER counter;
} CELL;

typedef struct {
	size_t off_help;
	size_t off_names;
	size_t celln;
	CELL cellv[1];
} VALUE;

static inline char *
cell_name(VALUE *vp, CELL *cp)
{
	return (char*)vp + vp->off_names + cp->off_name;
}

static inline char *
value_help(VALUE *vp)
{
	return (char*)vp + vp->off_help;
}

static CELL *
value_find_cell(VALUE *vp, char const *name)
{
	size_t i;
	for (i = 0; i < vp->celln; i++) {
		if (strcmp(cell_name(vp, &vp->cellv[i]), name) == 0)
			return &vp->cellv[i];
	}
	return NULL;
}

static CELL *
value_find_or_alloc(datum *dp, char const *name, char const *help)
{
	CELL *cp;

	if (!name)
		name = "";

	if (dp->dptr == NULL ||
	    (cp = value_find_cell((VALUE*)dp->dptr, name)) == NULL) {
		VALUE *vp;
		size_t namelen = strlen(name);
		size_t size;

		if (dp->dptr == NULL) {
			if (!help)
				return NULL;
			size = sizeof(VALUE) + namelen + strlen(help) + 2;
			vp = mu_alloc(size);
			memset(vp, 0, size);
			vp->off_help = sizeof(VALUE);
			strcpy(value_help(vp), help);
			vp->off_names = vp->off_help + strlen(help) + 1;
		} else {
			size = dp->dsize + sizeof(CELL) + namelen + 1;
			vp = mu_realloc(dp->dptr, size);
			memmove(vp->cellv + vp->celln + 1,
				vp->cellv + vp->celln,
				dp->dsize - vp->off_help);
			vp->off_help += sizeof(CELL);
			vp->off_names += sizeof(CELL);
		}
		cp = &vp->cellv[vp->celln];
		vp->celln++;
		cp->off_name = size - vp->off_names - namelen - 1;
		strcpy(cell_name(vp, cp), name);
		cp->counter = 0;
		dp->dptr = (char*)vp;
		dp->dsize = size;
	}
	return cp;
}


#define DEFAULT_ADDRESS "127.0.0.1"
#define DEFAULT_SERVICE "8080"
#define DEFAULT_TRUSTED_IP "127.0.0.0/8"

static char *dbname;
static struct MHD_Daemon *mhd;

static char *listen_address = DEFAULT_ADDRESS ":" DEFAULT_SERVICE;
static int httpd_access_log;
static mu_list_t trusted_ip_list;
static char *namespace;
static char *config_file = SYSCONFDIR "/mfmod_openmetrics.conf";

static int
cidr_comp(const void *a, const void *b)
{
	struct mu_cidr *ca = (struct mu_cidr *)a;
	struct mu_cidr *cb = (struct mu_cidr *)b;
	return mu_cidr_match(ca, cb);
}

static void
fill_trusted_ip_list(char *val)
{
       char *p;
       for (p = strtok(val, ","); p; p = strtok(NULL, ",")) {
	       struct mu_cidr cidr;
	       if (mu_cidr_from_string(&cidr, p) == 0) {
		       struct mu_cidr *cp = mu_alloc(sizeof(*cp));
		       memcpy(cp, &cidr, sizeof(*cp));
		       mu_list_append(trusted_ip_list, cp);
	       }
       }
}

/* Configuration */
static int
cb_trusted_ip(void *data, mu_config_value_t *val)
{
	struct mu_cidr cidr, *cp;
	int i;

	if (trusted_ip_list) {
		mu_list_create(&trusted_ip_list);
		mu_list_set_comparator(trusted_ip_list, cidr_comp);
	}

	switch (val->type) {
	case MU_CFG_STRING:
		if (mu_cidr_from_string(&cidr, val->v.string)) {
			mu_error("not a CIDR: %s", val->v.string);
			return 1;
		} else {
			cp = mu_alloc(sizeof(*cp));
			memcpy(cp, &cidr, sizeof(*cp));
			mu_list_append(trusted_ip_list, cp);
		}
		break;

	case MU_CFG_LIST:
	case MU_CFG_ARRAY:
		for (i = 0; i < val->v.arg.c; i++) {
			if (mu_cfg_assert_value_type(&val->v.arg.v[i], MU_CFG_STRING))
				return 1;
			if (mu_cidr_from_string(&cidr, val->v.arg.v[i].v.string)) {
				mu_error("not a CIDR: %s", val->v.arg.v[i].v.string);
				return 1;
			} else {
				cp = mu_alloc(sizeof(*cp));
				memcpy(cp, &cidr, sizeof(*cp));
				mu_list_append(trusted_ip_list, cp);
			}
		}
		break;
	}
	return 0;
}

static struct mu_cfg_param openmetrics_param[] = {
	{
		.ident = "listen",
		.type = mu_c_string,
		.data = &listen_address,
		.docstring = "Listen on this address.",
		.argname = "addr: [IP:]PORT"
	},
	{
		.ident = "access-log",
		.type = mu_c_bool,
		.data = &httpd_access_log,
		.docstring = "Enable HTTP access log.",
	},
	{
		.ident = "trusted-ip",
		.type = mu_cfg_callback,
		.callback = cb_trusted_ip,
		.docstring = "Trusted IP addresses."
	},
	{
		.ident = "namespace",
		.type = mu_c_string,
		.docstring = "Namespace prefix for metric names.",
		.data = &namespace
	},
	{ NULL }
};

static void
conf_read(char const *filename)
{
	mu_cfg_tree_t *tree;

	if (mu_cfg_parse_file(&tree, filename, 0))
		return;
	mu_cfg_tree_reduce(tree, NULL, openmetrics_param, NULL);
	mu_cfg_destroy_tree(&tree);
}

static int
openmetrics_dbopen(GDBM_FILE *ret_dbf, MFMOD_PARAM *r)
{
	GDBM_FILE dbf;
	if (!dbname) {
		int fd;

#define TEMPLATE "mfopenmetrics.XXXXXX"
		char *namebuf;
		char *tmpdir = getenv("TMP");
		if (!tmpdir)
			tmpdir = "/tmp";

		namebuf = mu_alloc(strlen(tmpdir) + 1 + sizeof(TEMPLATE));
		strcpy(namebuf, tmpdir);
		strcat(namebuf, "/");
		strcat(namebuf, TEMPLATE);

		if ((fd = mkstemp(namebuf)) == -1) {
			free(namebuf);
			r->type = mfmod_string;
			r->string = strerror(errno);
			return mfe_failure;
		}

		dbf = gdbm_fd_open(fd, namebuf, 0, GDBM_NEWDB, NULL);

		if (dbf == NULL) {
			free(namebuf);
			close(fd);
			return egdbm(r);
		}
		dbname = namebuf;
	} else {
		dbf = gdbm_open(dbname, 0, GDBM_WRCREAT, 0640, NULL);

		if (dbf == NULL) {
			return egdbm(r);
		}
	}

	*ret_dbf = dbf;

	return 0;
}

static void
cleanup(void)
{
	if (dbname) {
		unlink(dbname);
		free(dbname);
		dbname = NULL;
	}
}

static void server_start(void);

int
openmetrics_init(long count, MFMOD_PARAM *p, MFMOD_PARAM *r)
{
	int rc;
	GDBM_FILE dbf;

	if (count != 0)
		return enumargs(r);

	if (access(config_file, F_OK) == 0) {
		conf_read(config_file);
	}

	if ((rc = openmetrics_dbopen(&dbf, r)) != 0)
		return rc;
	gdbm_close(dbf);

	server_start();
	r->type = mfmod_number;
	r->number = 0;
	return 0;
}

int
openmetrics_destroy(long count, MFMOD_PARAM *p, MFMOD_PARAM *r)
{
	if (count != 0)
		return enumargs(r);
	MHD_stop_daemon(mhd);
	cleanup();
	r->type = mfmod_number;
	r->number = 0;
	return 0;
}

int
openmetrics_declare(long count, MFMOD_PARAM *p, MFMOD_PARAM *r)
{
	GDBM_FILE dbf;
	datum key, content;
	int rc;

	if (count != 2)
		return enumargs(r);
	ASSERT_ARGTYPE(p, r, 0, mfmod_string);
	ASSERT_ARGTYPE(p, r, 1, mfmod_string);

	if ((rc = openmetrics_dbopen(&dbf, r)) != 0)
		return rc;

	key.dptr = p->string;
	key.dsize = strlen(key.dptr);

	content = gdbm_fetch(dbf, key);

	value_find_or_alloc(&content, "", p[1].string);

	rc = gdbm_store(dbf, key, content, GDBM_REPLACE);
	free(content.dptr);
	gdbm_close(dbf);
	if (rc)
		return egdbm(r);

	r->type = mfmod_number;
	r->number = 0;

	return 0;
}

/*
 * Arguments:
 *    p[0]  string     name of the openmetrics
 *    p[1]  string     qualifier
 */
int
openmetrics_incr(long count, MFMOD_PARAM *p, MFMOD_PARAM *r)
{
	GDBM_FILE dbf;
	datum key, content;
	int rc;
	CELL *cp;

	if (count != 2)
		return enumargs(r);
	ASSERT_ARGTYPE(p, r, 0, mfmod_string);
	ASSERT_ARGTYPE(p, r, 1, mfmod_string);

	if ((rc = openmetrics_dbopen(&dbf, r)) != 0)
		return rc;

	key.dptr = p->string;
	key.dsize = strlen(key.dptr);
	content = gdbm_fetch(dbf, key);
	if (content.dptr == NULL) {
		if (gdbm_errno == GDBM_ITEM_NOT_FOUND) {
			char buf[256];

			r->type = mfmod_string;
			snprintf(buf, sizeof(buf), "openmetrics %s not declared",
				p->string);
			r->string = mu_strdup(buf);
			return mfe_not_found;
		} else {
			return egdbm(r);
		}
	}

	cp = value_find_or_alloc(&content, p[1].string, NULL);

	cp->counter++;
	rc = gdbm_store(dbf, key, content, GDBM_REPLACE);
	free(content.dptr);
	gdbm_close(dbf);
	if (rc)
		return egdbm(r);

	r->type = mfmod_number;
	r->number = 0;

	return 0;
}

/*
 * Arguments:
 *    p[0]  string     name of the openmetrics
 *    p[1]  string     qualifier
 */
int
openmetrics_clear(long count, MFMOD_PARAM *p, MFMOD_PARAM *r)
{
	GDBM_FILE dbf;
	datum key, content;
	int rc;
	CELL *cp;

	if (count != 2)
		return enumargs(r);
	ASSERT_ARGTYPE(p, r, 0, mfmod_string);
	ASSERT_ARGTYPE(p, r, 1, mfmod_string);

	if ((rc = openmetrics_dbopen(&dbf, r)) != 0)
		return rc;

	key.dptr = p->string;
	key.dsize = strlen(key.dptr);
	content = gdbm_fetch(dbf, key);
	if (content.dptr == NULL) {
		if (gdbm_errno == GDBM_ITEM_NOT_FOUND) {
			char buf[256];

			r->type = mfmod_string;
			snprintf(buf, sizeof(buf), "openmetrics %s not declared",
				p->string);
			r->string = mu_strdup(buf);
			return mfe_not_found;
		} else {
			return egdbm(r);
		}
	}

	cp = value_find_or_alloc(&content, p[1].string, NULL);

	cp->counter = 0;
	rc = gdbm_store(dbf, key, content, GDBM_REPLACE);
	free(content.dptr);
	gdbm_close(dbf);
	if (rc)
		return egdbm(r);

	r->type = mfmod_number;
	r->number = 0;

	return 0;
}
/*
 * Server
 */
static int
reassemble_get_args(void *cls, enum MHD_ValueKind kind,
		    const char *key, const char *value)
{
	mu_opool_t pool = cls;
	mu_opool_append_char(pool, '&');
	mu_opool_appendz(pool, key);
	if (value) {
		mu_opool_append_char(pool, '=');
		mu_opool_appendz(pool, value);
	}
	return MHD_YES;
}

static int
is_trusted_ip(char const *ipstr)
{
	struct mu_cidr cidr;

	if (mu_cidr_from_string(&cidr, ipstr))
		return 0;
	return mu_list_locate(trusted_ip_list, &cidr, NULL) == 0;
}

static char *
scan_forwarded_header(char const *hdr)
{
	char const *end;
	char *buf = NULL;
	size_t bufsize = 0;

	end = hdr + strlen(hdr);
	while (end > hdr) {
		char const *p;
		size_t len, i, j;

		for (p = end - 1; p > hdr && *p != ','; p--)
			;
		len = end - p;
		if (len + 1 > bufsize) {
			char *newbuf = realloc(buf, len + 1);
			if (newbuf) {
				buf = newbuf;
				bufsize = len + 1;
			} else {
				free(buf);
				return NULL;
			}
		}

		i = 0;
		j = 0;
		if (*p == ',')
			j++;
		while (j < len && isspace(p[j]))
			j++;
		while (j < len) {
			if (isspace(p[j]))
				break;
			buf[i++] = p[j++];
		}
		buf[i] = 0;

		if (!is_trusted_ip(buf))
			return buf;

		end = p;
	}
	free(buf);
	return NULL;
}

char *
get_remote_ip(struct MHD_Connection *conn)
{
	union MHD_ConnectionInfo const *ci;
	char ipstr[NI_MAXHOST];
	char const *hdr;
	char *ret;

	hdr = MHD_lookup_connection_value(conn,
					  MHD_HEADER_KIND,
					  "x-forwarded-for");
	if (hdr && (ret = scan_forwarded_header(hdr)) != NULL)
		return ret;

	ci = MHD_get_connection_info(conn,
				     MHD_CONNECTION_INFO_CLIENT_ADDRESS,
				     NULL);
	if (!ci)
		return NULL;

	if (getnameinfo(ci->client_addr, sizeof(struct sockaddr),
			ipstr, sizeof(ipstr), NULL, 0,
			NI_NUMERICHOST))
		return NULL;
	return strdup(ipstr);
}

static void
http_log(struct MHD_Connection *connection,
	 char const *method, char const *url,
	 int status, size_t size)
{
	char *ipstr;
	char const *host, *referer, *user_agent;
	time_t t;
	struct tm *tm;
	char tbuf[30];
	mu_opool_t pool;
	char *reqs = NULL;

	if (!httpd_access_log)
		return;
	ipstr = get_remote_ip(connection);

	host = MHD_lookup_connection_value(connection,
					   MHD_HEADER_KIND,
					   MHD_HTTP_HEADER_HOST);
	referer = MHD_lookup_connection_value(connection,
					      MHD_HEADER_KIND,
					      MHD_HTTP_HEADER_REFERER);
	user_agent = MHD_lookup_connection_value(connection,
						 MHD_HEADER_KIND,
						 MHD_HTTP_HEADER_USER_AGENT);
	t = time(NULL);
	tm = localtime(&t);
	strftime(tbuf, sizeof(tbuf), "[%d/%b/%Y:%H:%M:%S %z]", tm);

	mu_opool_create(&pool, MU_OPOOL_ENOMEMABRT);
	mu_opool_append_char(pool, '&');
	MHD_get_connection_values(connection, MHD_GET_ARGUMENT_KIND,
				  &reassemble_get_args, pool);
	if (mu_opool_size(pool) > 1) {
		mu_opool_append_char(pool, 0);
		reqs = mu_opool_finish(pool, NULL);
	}

	//FIXME
	if (size > 0)
		mu_error("%s %s - - %s \"%s %s%s\" %3d %zu \"%s\" \"%s\"",
			 host, ipstr, tbuf, method, url, reqs ? reqs : "",
			 status, size,
			 referer ? referer : "-",
			 user_agent ? user_agent : "-");
	else
		mu_error("%s %s - - %s \"%s %s%s\" %3d - \"%s\" \"%s\"",
			 host, ipstr, tbuf, method, url, reqs ? reqs : "",
			 status,
			 referer ? referer : "-",
			 user_agent ? user_agent : "-");

	mu_opool_destroy(&pool);
	free(ipstr);
}

struct openmetrics_iterator
{
	GDBM_FILE dbf;
	datum key;
	datum content;
	VALUE *val;
	size_t cell;
	char *buf;
	size_t bufsize;
	size_t buflen;
	size_t bufpos;
	int eof;
	size_t bytes_sent;

	struct MHD_Connection *conn;
	char const *method;
	char const *url;
};

static struct openmetrics_iterator *
openmetrics_iterator_init(struct MHD_Connection *conn, char const *method, char const *url)
{
	struct openmetrics_iterator *itr = mu_alloc(sizeof(*itr));
	memset(itr, 0, sizeof(*itr));

	itr->dbf = gdbm_open(dbname, 0, GDBM_READER, 0, NULL);
	if (!itr->dbf) {
		mu_error("can't open database %s for reading: %s (%s)",
			 dbname, gdbm_strerror(gdbm_errno), strerror(errno));
		free(itr);
		return NULL;
	}

	itr->conn = conn;
	itr->method = method;
	itr->url = url;
	return itr;
}

static ssize_t
movebuf(struct openmetrics_iterator *itr, char **buf, size_t *max)
{
	size_t n = itr->buflen - itr->bufpos;
	if (n > 0) {
		if (n > *max)
			n = *max;

		memcpy(*buf, itr->buf + itr->bufpos, n);
		itr->bufpos += n;
		itr->bytes_sent += n;
		*buf += n;
		*max -= n;
	}
	return n;
}

static ssize_t
response_reader_callback(void *cls, uint64_t pos, char *buf, size_t max)
{
	struct openmetrics_iterator *itr = cls;
	CELL *cell;
	char const *name;
	ssize_t total = 0;
	char *namespace_prefix;
	char *namespace_delim;

	if (namespace) {
		namespace_prefix = namespace;
		namespace_delim = "_";
	} else {
		namespace_prefix = "";
		namespace_delim = "";
	}
	
	while (1) {
		total += movebuf(itr, &buf, &max);
		if (max == 0)
			break;
		if (itr->eof) {
			if (total > 0)
				break;
			http_log(itr->conn, itr->method, itr->url, MHD_HTTP_OK, itr->bytes_sent);
			return MHD_CONTENT_READER_END_OF_STREAM;
		}

		if (!itr->val || itr->cell == itr->val->celln) {
			datum key;

			free(itr->content.dptr);
			itr->content.dptr = NULL;
			itr->val = NULL;

			if (itr->key.dptr == NULL)
				key = gdbm_firstkey(itr->dbf);
			else
				key = gdbm_nextkey(itr->dbf, itr->key);
			free(itr->key.dptr);
			itr->key = key;
			if (key.dptr == NULL) {
				itr->eof = 1;
				mu_asnprintf(&itr->buf, &itr->bufsize, "# EOF\n");
				itr->buflen = strlen(itr->buf);
				itr->bufpos = 0;
				continue;
			}
			itr->content = gdbm_fetch(itr->dbf, itr->key);
			if (itr->content.dptr == NULL) {
				mu_error("can't get value for openmetrics %*.*s: %s",
					 itr->key.dsize, itr->key.dsize, itr->key.dptr,
					 gdbm_strerror(gdbm_errno));
				return MHD_CONTENT_READER_END_WITH_ERROR;
			}

			itr->val = (VALUE*) itr->content.dptr;
			itr->cell = 0;

			mu_asnprintf(&itr->buf, &itr->bufsize,
				     "# HELP %*.*s %s\n# TYPE %*.*s counter\n",
				     itr->key.dsize, itr->key.dsize, itr->key.dptr,
				     value_help(itr->val),
				     itr->key.dsize, itr->key.dsize, itr->key.dptr);
			itr->buflen = strlen(itr->buf);
			itr->bufpos = 0;
			continue;
		}

		cell = &itr->val->cellv[itr->cell++];
		name = cell_name(itr->val, cell);
		if (name[0]) {
			mu_asnprintf(&itr->buf, &itr->bufsize,
				     "%s%s%*.*s{%s} %"PRIu64"\n",
				     namespace_prefix, namespace_delim,
				     itr->key.dsize, itr->key.dsize, itr->key.dptr,
				     name,
				     cell->counter);
		} else {
			mu_asnprintf(&itr->buf, &itr->bufsize,
				     "%s%s%*.*s %"PRIu64"\n",
				     namespace_prefix, namespace_delim,
				     itr->key.dsize, itr->key.dsize, itr->key.dptr,
				     cell->counter);
		}
		itr->buflen = strlen(itr->buf);
		itr->bufpos = 0;
	}

	return total;
}

static void
response_free_callback(void *cls)
{
	struct openmetrics_iterator *itr = cls;
	free(itr->key.dptr);
	free(itr->content.dptr);
	free(itr->buf);
	gdbm_close(itr->dbf);
	free(itr);
}

static int
http_response(struct MHD_Connection *conn,
	      char const *method, char const *url, int status)
{
	int ret;
	struct MHD_Response *resp =
		MHD_create_response_from_buffer(0,
						NULL,
						MHD_RESPMEM_PERSISTENT);
	http_log(conn, method, url, status, 0);
	ret = MHD_queue_response(conn, status, resp);
	MHD_destroy_response(resp);
	return ret;
}

static int
openmetrics_httpd_handler(void *cls,
		      struct MHD_Connection *conn,
		      const char *url, const char *method,
		      const char *version,
		      const char *upload_data, size_t *upload_data_size,
		      void **con_cls)
{
	struct MHD_Response *response;
	int ret;
	struct openmetrics_iterator *itr;

	if (strcmp(method, MHD_HTTP_METHOD_GET) &&
	    strcmp(method, MHD_HTTP_METHOD_HEAD))
		return http_response(conn, method, url,
				     MHD_HTTP_METHOD_NOT_ALLOWED);

	if (strcmp(url, "/metrics")) {
		mu_error("bad url: %s", url);
		return http_response(conn, method, url, MHD_HTTP_NOT_FOUND);
	}

	if (!dbname) {
		mu_error("%s", "no database name");
		return http_response(conn, method, url,
				     MHD_HTTP_INTERNAL_SERVER_ERROR);
	}

	if ((itr = openmetrics_iterator_init(conn, method, url)) == NULL)
		return http_response(conn, method, url,
				     MHD_HTTP_INTERNAL_SERVER_ERROR);

	response = MHD_create_response_from_callback(-1, 4096,
						     response_reader_callback,
						     itr,
						     response_free_callback);
	MHD_add_response_header(response,
				MHD_HTTP_HEADER_CONTENT_TYPE,
				"text/plain; version=0.0.4; charset=utf-8");

	ret = MHD_queue_response(conn, MHD_HTTP_OK, response);
	MHD_destroy_response(response);
	return ret;
}

static void
openmetrics_httpd_logger(void *arg, const char *fmt, va_list ap)
{
	mu_verror(fmt, ap);
}

static void
openmetrics_httpd_panic(void *cls, const char *file, unsigned int line,
		    const char *reason)
{
	if (reason)
		mu_error("%s:%d: MHD PANIC: %s", file, line, reason);
	else
		mu_error("%s:%d: MHD PANIC", file, line);
	abort();
}

static int
open_node(char const *node, char const *serv, struct sockaddr **saddr)
{
	struct addrinfo hints;
	struct addrinfo *res;
	int reuse;
	int fd;
	int rc;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	rc = getaddrinfo(node, serv, &hints, &res);
	if (rc) {
		mu_error("%s: %s", node, gai_strerror(rc));
		exit(1);
	}

	fd = socket(res->ai_family, res->ai_socktype, 0);
	if (fd == -1) {
		mu_error("socket: %s", strerror(errno));
		return -1;
	}
	reuse = 1;
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
		   &reuse, sizeof(reuse));
	if (bind(fd, res->ai_addr, res->ai_addrlen) == -1) {
		mu_error("bind: %s", strerror(errno));
		return -1;
	}

	if (listen(fd, 8) == -1) {
		mu_error("listen: %s", strerror(errno));
		return -1;
	}

	if (saddr) {
		*saddr = mu_alloc(res->ai_addrlen);
		memcpy(*saddr, res->ai_addr, res->ai_addrlen);
	}
	freeaddrinfo(res);

	return fd;
}

static int
open_listener(char const *addr, struct sockaddr **saddr)
{
	size_t len;
	int fd;

	if (!addr)
		return open_node(DEFAULT_ADDRESS, DEFAULT_SERVICE, saddr);

	len = strcspn(addr, ":");
	if (len == 0)
		fd = open_node(DEFAULT_ADDRESS, addr + 1, saddr);
	else if (addr[len] == 0)
		fd = open_node(addr, DEFAULT_SERVICE, saddr);
	else {
		char *node;
		node = mu_alloc(len + 1);
		memcpy(node, addr, len);
		node[len] = 0;
		fd = open_node(node, addr + len + 1, saddr);
		free(node);
	}
	return fd;
}

static void
server_start(void)
{
	int fd;

	fd = open_listener(listen_address, NULL);
	if (fd == -1)
		return;

	if (!trusted_ip_list) {
		char *val = strdup(DEFAULT_TRUSTED_IP);
		fill_trusted_ip_list(val);
		free(val);
	}

	MHD_set_panic_func(openmetrics_httpd_panic, NULL);
	mhd = MHD_start_daemon(MHD_USE_INTERNAL_POLLING_THREAD
			       | MHD_USE_ERROR_LOG,
			       0,
			       NULL, NULL,
			       openmetrics_httpd_handler, NULL,
			       MHD_OPTION_LISTEN_SOCKET, fd,
			       MHD_OPTION_EXTERNAL_LOGGER, openmetrics_httpd_logger, NULL,
			       MHD_OPTION_END);
}
